#ifndef include_omf_h
#define include_omf_h

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

typedef unsigned char byte;

struct PString
{
  byte count;
  char text[0];
};

static inline PString *Pstrdup (const PString *s)
{
  PString *r = (PString *) malloc (s->count + 1);
  memcpy (r, s, s->count + 1);
  return r;
}

static inline void pstring (const PString *string)
{
  fwrite (string->text, string->count, 1, stdout);
}

class OMFProcessor
{
private:
  friend int OMFProcess (FILE *f, OMFProcessor *proc);
  int index_alloc_size;
  int index_size;
  PString **index;
  int parse_stop;
protected:
  void append_index_name (const PString *name);
  void omf_parse_stop (void) { parse_stop = 1; }
public:
  int omf_error;
  OMFProcessor ();
  virtual ~OMFProcessor ();

  const PString *lookup_index_name (unsigned short index);

  virtual void raw_record (byte type, byte *data, unsigned short len);
  virtual void unknown_record (byte type, byte *data, unsigned short len);
  void format_error (const char *err = 0, ...);
  virtual void va_format_error (const char *err, va_list va);
  
  virtual void raw_theadr (byte *data, unsigned short len);
  virtual void theadr (PString *string);
  virtual void raw_lheadr (byte *data, unsigned short len);
  virtual void raw_coment (byte *data, unsigned short len);
  virtual void coment (int nopurge, int nolist, int coment_class,
			byte *data, unsigned short len);
  virtual void coment_translator (byte *name, unsigned len);
  virtual void coment_default_library (byte *name, unsigned len);
  virtual void coment_end_of_link_pass_one ();
  virtual void coment_other (int coment_class, byte *data, unsigned short len);
  virtual void raw_modend (byte *data, unsigned short len);
  virtual void raw_modend32 (byte *data, unsigned short len);
  virtual void raw_extdef (byte *data, unsigned short len);
  virtual void extdef (PString *name, unsigned short type);
  virtual void raw_pubdef (byte *data, unsigned short len);
  virtual void raw_pubdef32 (byte *data, unsigned short len);
  virtual void pubdef32 (unsigned short group, unsigned short segment,
			 PString *name, unsigned long offset,
			 unsigned short type);
  virtual void raw_linnum (byte *data, unsigned short len);
  virtual void raw_linnum32 (byte *data, unsigned short len);
  virtual void raw_lnames (byte *data, unsigned short len);
  virtual void lname (PString *name);
  virtual void raw_segdef (byte *data, unsigned short len);
  virtual void raw_segdef32 (byte *data, unsigned short len);
  virtual void segdef32 (byte alignment, byte combination, byte big,
			 byte bit32, unsigned long segment_length,
			 const PString *segment_name,
			 const PString *class_name,
			 const PString *overlay_name);
  virtual void raw_grpdef (byte *data, unsigned short len);
  virtual void raw_fixupp (byte *data, unsigned short len);
  virtual void raw_fixupp32 (byte *data, unsigned short len);
  virtual void raw_ledata (byte *data, unsigned short len);
  virtual void raw_ledata32 (byte *data, unsigned short len);
  virtual void raw_lidata (byte *data, unsigned short len);
  virtual void raw_lidata32 (byte *data, unsigned short len);
  virtual void raw_comdef (byte *data, unsigned short len);
  virtual void comdef (PString *name, unsigned short type,
		       unsigned long length);
  virtual void raw_bakpat (byte *data, unsigned short len);
  virtual void raw_bakpat32 (byte *data, unsigned short len);
  virtual void raw_lextdef (byte *data, unsigned short len);
  virtual void lextdef (PString *name, unsigned short type);
  virtual void raw_lpubdef (byte *data, unsigned short len);
  virtual void raw_lpubdef32 (byte *data, unsigned short len);
  virtual void lpubdef32 (unsigned short group, unsigned short segment,
			  PString *name, unsigned long offset,
			  unsigned short type);
  virtual void raw_lcomdef (byte *data, unsigned short len);
  virtual void lcomdef (PString *name, unsigned short type,
			unsigned long length);
  virtual void raw_cextdef (byte *data, unsigned short len);
  virtual void raw_comdat (byte *data, unsigned short len);
  virtual void raw_comdat32 (byte *data, unsigned short len);
  virtual void raw_linsym (byte *data, unsigned short len);
  virtual void raw_linsym32 (byte *data, unsigned short len);
  virtual void raw_alias (byte *data, unsigned short len);
  virtual void raw_nbkpat (byte *data, unsigned short len);
  virtual void raw_nbkpat32 (byte *data, unsigned short len);
  virtual void raw_llnames (byte *data, unsigned short len);
  virtual void raw_vernum (byte *data, unsigned short len);
  virtual void raw_vendext (byte *data, unsigned short len);
  virtual void raw_library_header_record (byte *data, unsigned short len);
  virtual void raw_library_end_record (byte *data, unsigned short len);
};

int OMFProcess (FILE *f, OMFProcessor *);

static inline unsigned short extract_omf_short (const byte *omf)
{
  unsigned short lo = omf[0];
  unsigned short hi = omf[1];
  return hi * 256 + lo;
}

static inline unsigned long extract_omf_long (const byte *omf)
{
  unsigned long lo = extract_omf_short (omf);
  unsigned long hi = extract_omf_short (omf + 2);
  return hi * (256 * 256) + lo;
}

static inline int omf_index_is_one_byte (const byte *omf)
{
  return !(omf[0] & 0x80);
}

static inline unsigned short extract_omf_index (const byte *omf, int *len)
{
  if (omf_index_is_one_byte (omf))
    {
      *len = 1;
      return omf[0];
    }
  else
    {
      *len = 2;
      return extract_omf_short (omf);
    }
}

#endif /* include_omf_h */
