#include <stdio.h>
#include <ctype.h>
#include "omf.h"

class OMFDump: public OMFProcessor
{
  typedef OMFProcessor inherited;
public:
  virtual void unknown_record (byte type, byte *data, unsigned short len);
  virtual void va_format_error (const char *err, va_list va);
  virtual void theadr (PString *string);
  virtual void raw_lheadr (byte *data, unsigned short len);
  virtual void coment (int nopurge, int nolist, int coment_class,
			byte *data, unsigned short len);
  virtual void coment_translator (byte *name, unsigned len);
  virtual void coment_default_library (byte *name, unsigned len);
  virtual void coment_end_of_link_pass_one ();
  virtual void coment_other (int coment_class, byte *data, unsigned short len);
  virtual void raw_modend (byte *data, unsigned short len);
  virtual void raw_modend32 (byte *data, unsigned short len);
  virtual void extdef (PString *name, unsigned short type);
  virtual void raw_pubdef (byte *data, unsigned short len);
  void pubdef32 (unsigned short group, unsigned short segment,
		 PString *name, unsigned long offset,
		 unsigned short type);
  virtual void raw_linnum (byte *data, unsigned short len);
  virtual void raw_linnum32 (byte *data, unsigned short len);
  virtual void raw_lnames (byte *data, unsigned short len);
  virtual void lname (PString *name);
  virtual void raw_segdef (byte *data, unsigned short len);
  virtual void segdef32 (byte alignment, byte combination, byte big,
			 byte bit32, unsigned long segment_length,
			 const PString *segment_name,
			 const PString *class_name,
			 const PString *overlay_name);
  virtual void raw_grpdef (byte *data, unsigned short len);
  virtual void raw_fixupp (byte *data, unsigned short len);
  virtual void raw_fixupp32 (byte *data, unsigned short len);
  virtual void raw_ledata (byte *data, unsigned short len);
  virtual void raw_ledata32 (byte *data, unsigned short len);
  virtual void raw_lidata (byte *data, unsigned short len);
  virtual void raw_lidata32 (byte *data, unsigned short len);
  virtual void comdef (PString *name, unsigned short type,
		       unsigned long length);
  virtual void raw_bakpat (byte *data, unsigned short len);
  virtual void raw_bakpat32 (byte *data, unsigned short len);
  virtual void lextdef (PString *name, unsigned short type);
  virtual void raw_lpubdef (byte *data, unsigned short len);
  virtual void lpubdef32 (unsigned short group, unsigned short segment,
			  PString *name, unsigned long offset,
			  unsigned short type);
  virtual void lcomdef (PString *name, unsigned short type,
			unsigned long length);
  virtual void raw_cextdef (byte *data, unsigned short len);
  virtual void raw_comdat (byte *data, unsigned short len);
  virtual void raw_comdat32 (byte *data, unsigned short len);
  virtual void raw_linsym (byte *data, unsigned short len);
  virtual void raw_linsym32 (byte *data, unsigned short len);
  virtual void raw_alias (byte *data, unsigned short len);
  virtual void raw_nbkpat (byte *data, unsigned short len);
  virtual void raw_nbkpat32 (byte *data, unsigned short len);
  virtual void raw_llnames (byte *data, unsigned short len);
  virtual void raw_vernum (byte *data, unsigned short len);
  virtual void raw_vendext (byte *data, unsigned short len);
  virtual void raw_library_header_record (byte *data, unsigned short len);
  virtual void raw_library_end_record (byte *data, unsigned short len);
} omf_dumper;

void dump_raw_record (const char *type, byte *data, unsigned short len)
{
  if (type)
    printf ("%s: data length %d (0x%02X)\n", type, len, len);
  int i, k;
  for (i = 0; i < len; i += 16)
    {
      printf ("  %04X ", i);
      for (k = 0; k < 8; k++)
	if (i + k < len)
	  printf (" %02X", data[i+k]);
	else
	  printf ("   ");
      printf (" ");
      for (k = 0; k < 8; k++)
	if (i + k + 8 < len)
	  printf (" %02X", data[i+k+8]);
	else
	  printf ("   ");
      printf ("  ");
      for (k = 0; k < 8; k++)
	if (i + k < len)
	  if (isprint (data[i+k]))
	    printf ("%c", data[i+k]);
	  else
	    printf (".");
	else
	  printf (" ");
      printf (" ");
      for (k = 0; k < 8; k++)
	if (i + k + 8 < len)
	  if (isprint (data[i+k+8]))
	    printf ("%c", data[i+k+8]);
	  else
	    printf (".");
	else
	  printf (" ");
      printf ("\n");
    }
}

void OMFDump::unknown_record (byte type, byte *data, unsigned short len)
{
  printf ("UNKNOWN RECORD TYPE 0x%02X: data length %d (0x%02X)\n",
	  type, len, len);
  dump_raw_record (0, data, len);
}

void OMFDump::va_format_error (const char *err, va_list va)
{
  if (err == 0)
    printf ("Object module format error\n");
  else
    {
      printf ("Object module format error:\n");
      vprintf (err, va);
      printf ("\n");
    }
  inherited::va_format_error (err, va);
}

void OMFDump::theadr (PString *string)
{
  printf ("THEADR: '");
  pstring (string);
  printf ("'\n");
}

void OMFDump::raw_lheadr (byte *data, unsigned short len)
{
  dump_raw_record ("LHEADR", data, len);
}

void OMFDump::coment (int nopurge, int nolist, int coment_class,
	     byte *data, unsigned short len)
{
  printf ("COMENT: %spurge, %slist",
	  nopurge ? "no " : "",
	  nolist ? "no " : "");
  inherited::coment (nopurge, nolist, coment_class, data, len);
}

void OMFDump::coment_translator (byte *name, unsigned len)
{
  printf ("\n  Translator: '");
  fwrite (name, len, 1, stdout);
  printf ("'\n");
}

void OMFDump::coment_default_library (byte *name, unsigned len)
{
  printf ("\n  Default library: '");
  fwrite (name, len, 1, stdout);
  printf ("'\n");
}

void OMFDump::coment_end_of_link_pass_one ()
{
  printf ("\n  End of link pass 1\n");
}

void OMFDump::coment_other (int coment_class, byte *data, unsigned short len)
{
  printf (", class %d (0x%02X)\n", coment_class, coment_class);
  dump_raw_record (0, data, len);
}

void OMFDump::raw_modend (byte *data, unsigned short len)
{
  dump_raw_record ("MODEND", data, len);
}

void OMFDump::raw_modend32 (byte *data, unsigned short len)
{
  dump_raw_record ("MODEND32", data, len);
}

void OMFDump::extdef (PString *s, unsigned short type)
{
  printf ("EXTDEF: '");
  pstring (s);
  printf ("' type: 0x%X\n", type);
}

void OMFDump::raw_pubdef (byte *data, unsigned short len)
{
  dump_raw_record ("PUBDEF", data, len);
}

void OMFDump::pubdef32 (unsigned short group, unsigned short segment,
			PString *name, unsigned long offset,
			unsigned short type)
{
  printf ("PUBDEF32: group:%d seg:%d ofs:0x%lX type:0x%X '",
	  group, segment, offset, type);
  pstring (name);
  printf ("'\n");
}
			 
void OMFDump::raw_linnum (byte *data, unsigned short len)
{
  dump_raw_record ("LINNUM", data, len);
}

void OMFDump::raw_linnum32 (byte *data, unsigned short len)
{
  dump_raw_record ("LINNUM32", data, len);
}

void OMFDump::raw_lnames (byte *data, unsigned short len)
{
  printf ("LNAMES:\n");
  inherited::raw_lnames (data, len);
}

void OMFDump::lname (PString *string)
{
  printf ("  '");
  pstring (string);
  printf ("'\n");
  inherited::lname (string);
}

void OMFDump::raw_segdef (byte *data, unsigned short len)
{
  dump_raw_record ("SEGDEF", data, len);
}

void OMFDump::segdef32 (byte alignment, byte combination, byte big,
			byte use32, unsigned long segment_length,
			const PString *segment_name, const PString *class_name,
			const PString *overlay_name)
{
  printf ("SEGDEF32: align:");
  switch (alignment)
    {
    case 0:
      printf ("absolute"); break;
    case 1:
      printf ("byte"); break;
    case 2:
      printf ("word"); break;
    case 3:
      printf ("para"); break;
    case 4:
      printf ("page"); break;
    case 5:
      printf ("dword"); break;
    default:
      printf ("%d", alignment); break;
    }
  printf (" comb:");
  switch (combination)
    {
    case 0:
      printf ("private, no combine"); break;
    case 2:
    case 4:
    case 7:
      printf ("public, combine aligned"); break;
    case 5:
      printf ("stack"); break;
    case 6:
      printf ("common"); break;
    default:
      printf ("%d", combination); break;
    }
  if (use32)
    printf (" use32 ");
  else
    printf (" use16 ");
  if (big)
    printf ("big ");
  printf ("len: %ld (0x%lX)\n", segment_length, segment_length);
  printf ("  Segment name: '");
  pstring (segment_name);
  printf ("'\n  Class name: '");
  pstring (class_name);
  printf ("'\n  Overlay name: '");
  pstring (overlay_name);
  printf ("'\n");
}

void OMFDump::raw_grpdef (byte *data, unsigned short len)
{
  dump_raw_record ("GRPDEF", data, len);
}

void OMFDump::raw_fixupp (byte *data, unsigned short len)
{
  dump_raw_record ("FIXUPP", data, len);
}

void OMFDump::raw_fixupp32 (byte *data, unsigned short len)
{
  dump_raw_record ("FIXUPP32", data, len);
}

void OMFDump::raw_ledata (byte *data, unsigned short len)
{
  dump_raw_record ("LEDATA", data, len);
}

void OMFDump::raw_ledata32 (byte *data, unsigned short len)
{
  dump_raw_record ("LEDATA32", data, len);
}

void OMFDump::raw_lidata (byte *data, unsigned short len)
{
  dump_raw_record ("LIDATA", data, len);
}

void OMFDump::raw_lidata32 (byte *data, unsigned short len)
{
  dump_raw_record ("LIDATA32", data, len);
}

void OMFDump::comdef (PString *name, unsigned short type,
		      unsigned long length)
{
  printf ("COMDEF: type:%d length:%ld (0x%lX) '", type, length, length);
  pstring (name);
  printf ("'\n");
}

void OMFDump::raw_bakpat (byte *data, unsigned short len)
{
  dump_raw_record ("BAKPAT", data, len);
}

void OMFDump::raw_bakpat32 (byte *data, unsigned short len)
{
  dump_raw_record ("BAKPAT32", data, len);
}

void OMFDump::lextdef (PString *s, unsigned short type)
{
  printf ("LEXTDEF: '");
  pstring (s);
  printf ("' type: 0x%X\n", type);
}

void OMFDump::raw_lpubdef (byte *data, unsigned short len)
{
  dump_raw_record ("LPUBDEF", data, len);
}

void OMFDump::lpubdef32 (unsigned short group, unsigned short segment,
			PString *name, unsigned long offset,
			unsigned short type)
{
  printf ("LPUBDEF32: group:%d seg:%d ofs:0x%lX type:0x%X '",
	  group, segment, offset, type);
  pstring (name);
  printf ("'\n");
}

void OMFDump::lcomdef (PString *name, unsigned short type,
		       unsigned long length)
{
  printf ("LCOMDEF: type:%d length:%ld (0x%lX) '", type, length, length);
  pstring (name);
  printf ("'\n");
}

void OMFDump::raw_cextdef (byte *data, unsigned short len)
{
  dump_raw_record ("CEXTDEF", data, len);
}

void OMFDump::raw_comdat (byte *data, unsigned short len)
{
  dump_raw_record ("COMDAT", data, len);
}

void OMFDump::raw_comdat32 (byte *data, unsigned short len)
{
  dump_raw_record ("COMDAT32", data, len);
}

void OMFDump::raw_linsym (byte *data, unsigned short len)
{
  dump_raw_record ("LINSYM", data, len);
}

void OMFDump::raw_linsym32 (byte *data, unsigned short len)
{
  dump_raw_record ("LINSYM32", data, len);
}

void OMFDump::raw_alias (byte *data, unsigned short len)
{
  dump_raw_record ("ALIAS", data, len);
}

void OMFDump::raw_nbkpat (byte *data, unsigned short len)
{
  dump_raw_record ("NBKPAT", data, len);
}

void OMFDump::raw_nbkpat32 (byte *data, unsigned short len)
{
  dump_raw_record ("NBKPAT32", data, len);
}

void OMFDump::raw_llnames (byte *data, unsigned short len)
{
  dump_raw_record ("LLNAMES", data, len);
}

void OMFDump::raw_vernum (byte *data, unsigned short len)
{
  dump_raw_record ("VERNUM", data, len);
}

void OMFDump::raw_vendext (byte *data, unsigned short len)
{
  dump_raw_record ("VENDEXT", data, len);
}

void OMFDump::raw_library_header_record (byte *data, unsigned short len)
{
  dump_raw_record ("LIBRARY HEADER RECORD", data, len);
}

void OMFDump::raw_library_end_record (byte *data, unsigned short len)
{
  dump_raw_record ("LIBRARY END RECORD", data, len);
}


void omfdump (const char *filename)
{
  FILE *f = fopen (filename, "rb");
  if (f == 0)
    {
      fprintf (stderr, "omfdump: Unable to open file '%s'\n", filename);
      return;
    }
  printf ("Object Module Format dump of file '%s'\n", filename);
  OMFProcess (f, &omf_dumper);
  fclose (f);
}

int main (int argc, char **argv)
{
  int i;
  for (i = 1; i < argc; i++)
    omfdump (argv[i]);
  return 0;
}
