#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include "omf.h"

class OMFExp: public OMFProcessor
{
  typedef OMFProcessor inherited;
  const char *filename;
private:
  static void print_export (PString *name);
  void va_format_error (const char *err, va_list va);
  virtual void pubdef32 (unsigned short group, unsigned short segment,
			 PString *name, unsigned long offset,
			 unsigned short type);
  virtual void comdef (PString *name, unsigned short type,
		       unsigned long length);
  virtual void coment_end_of_link_pass_one ();
  void format_error (const char *err = 0, ...);
public:
  void set_filename (const char *afilename) { filename = afilename; }
};

void OMFExp::print_export (PString *name)
{
  printf ("\t");
  pstring (name);
  printf ("\n");
}

void OMFExp::pubdef32 (unsigned short group, unsigned short segment,
		       PString *name, unsigned long offset,
		       unsigned short type)
{
  inherited::pubdef32 (group, segment, name, offset, type);
  print_export (name);
}

void OMFExp::comdef (PString *name, unsigned short type,
		     unsigned long length)
{
  inherited::comdef (name, type, length);
  print_export (name);
}

void OMFExp::coment_end_of_link_pass_one ()
{
  inherited::coment_end_of_link_pass_one ();
  omf_parse_stop ();
}

void OMFExp::va_format_error (const char *err, va_list va)
{
  fprintf (stderr, "omfexp: OMF format error in file '%s'\n", filename);
  if (err)
    {
      fprintf (stderr, "omfexp: ");
      vfprintf (stderr, err, va);
      fprintf (stderr, "\n");
    }
}

OMFExp omf_exp;

void omfexp (const char *filename)
{
  FILE *f = fopen (filename, "rb");
  if (f == 0)
    {
      fprintf (stderr, "omfexp: unable to open file '%s'\n", filename);
      exit (1);
    }
  omf_exp.set_filename (filename);
  if (!OMFProcess (f, &omf_exp))
    exit (1);
  fclose (f);
}


int main (int argc, char **argv)
{
  int i;
  static char buf[1024];
  setvbuf (stdout, buf, _IOFBF, sizeof buf);
  for (i = 1; i < argc; i++)
    omfexp (argv[i]);
  return 0;
}
